/* Parse printf format string.
   Copyright (C) 1999, 2002, 2005, 2007, 2009-2013 Free Software Foundation,
   Inc.

   This program is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as published
   by the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef _U_PRINTF_PARSE_H
#define _U_PRINTF_PARSE_H

#include "unistdio/u-printf-args.h"

/* Parametrization of printf-parse.h.  */
#undef ENABLE_UNISTDIO
#define ENABLE_UNISTDIO 1
#include "printf-parse.h"

#endif /* _U_PRINTF_PARSE_H */
